from vector import *


class Matrix(tuple):
    """
    2d-Matrix class for calculations. A 'Matrix' object is a 2-element list
    of vectors with 'self[0]' and 'self[1]' accessable as 'self.i' and 'self.j'.
    There are also numerus methods to calculate with vectors.
    """


    def __new__(cls, i, j):
        """
        Return a new tuple '(i, j)' for parent of new 'Matrix' object.
        """

        try:
            return tuple.__new__(cls, (Vector(*i), Vector(*j)))
        except:
            raise TypeError(f"'i' and 'j' of matrix have to be type 'Vector'")


    def __getattr__(self, attr):
        """
        Return 'self[0]' and 'self[1]' as attribute 'i' and 'j'.
        """

        if attr == 'i':
            return self[0]
        if attr == 'j':
            return self[1]

        raise AttributeError(f"{self.__class__.__name__} has no Attribute {attr}.")


    def __copy__(self):
        """Return a copy of 'self'."""
        return Matrix(*self)

    def __bool__(self):
        """Return 'True' if 'self' is nonzero."""
        return bool(self.i) or bool(self.j)


    def __eq__(self, other):
        """
        Return 'True' if 'self' and 'other' are equal.
        """

        if type(other) is Matrix:
            return self.i == other.i and self.j == other.j
        else:
            return False


    def __ne__(self, other):
        """Return 'True' if 'self' and 'other' are not equal."""
        return not self == other


    def __neg__(self):
        """Return the additive inverse of 'self'."""
        return Matrix(-self.i, -self.j)


    def __invert__(self):
        """Swiches 'i' and 'j'."""
        return Matrix(self.j, self.i)


    def __round__(self, d=None):
        """Return 'self' with both components rounded to 'd' decimal places."""
        return Vector(round(self.i, d), round(self.j, d))


    def __add__(self, other):
        """
        Return the sum of matrices 'self' and 'other'.
        """

        if type(other) is Matrix:
            return Matrix(self.i + other.i, self.j + other.j)

        if type(other) in (int, float):
            return Matrix(self.i + Vector(other, 0), self.j + Vector(0, other))

        return NotImplemented


    def __radd__(self, other):
        """Return the sum of matrices 'self' and 'other'."""
        return self + other


    def __sub__(self, other):
        """Return the difference of matrices 'self' and 'other'."""
        return self + (-other)


    def __mul__(self, other):
        """
        If other is a integer or float
        Return a scaled matrix of 'self' by factor 'other'.

        If other is a matrix:
        Multiplies both components individually.
        """

        if type(other) in (int, float):
            return Matrix(self.i * other, self.j * other)

        if type(other) is Vector:
            return Matrix(self.i * other.i, self.j * other.j)

        return NotImplemented


    def __rmul__(self, value):
        """Return a scaled vector of 'self' by factor 'value'."""
        return self * value


    def __truediv__(self, other):
        """
        If other is a integer or float:
        Return a scaled vector of 'self' by factor '1 / other'.

        If other is a vector:
        Return 'value' with 'other * value == self'.
        Return 'None' if both are linear independant.
        """

        if type(other) in (int, float):
            return self * (1 / other)

        if type(other) is Vector:
            value = 0

            if other.x != 0:
                value = self.x / other.x
            elif other.y != 0:
                value = self.y / other.y

            if round(other * value, 5) == round(self, 5):
                return value

            return None

        return NotImplemented


    def __str__(self):
        """Return a string representing 'self' in style of '"<i, j>"'."""
        return f"<{tuple.__repr__(self)[1:-1]}>"


    def __repr__(self):
        """Return a string representing 'self' in style of '"<i, j>"'."""
        return str(self)


    def T(self):
        """Return the transposed matrix of 'self'."""
        return Matrix((self.i.x, self.j.x), (self.i.y, self.j.y))


    def det(self):
        """Return determinant of self."""
        return self.i.x * self.j.y - self.i.y * self.j.x


    def inverse(self):
        """Return the inverse of self."""
        return Matrix((-self.j.y, self.i.y), (self.j.x, -self.i.x)) / -self.det()


    def prod(self, other):
        """
        Return the matrix multiplication of 'self' and 'other'.
        """

        if type(other) is Vector:
            return Vector(self.T().i.dot(other), self.T().j.dot(other))

        if type(other) is Matrix:
            return Matrix(self.prod(other.i), self.prod(other.j))

        return NotImplemented
